/*
 * Copyright (c) 2014 Lutz Kaemmerer, 2015 Toni Volkmer
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <complex.h>
#include <math.h>

#ifdef __linux__
  /* for getcurtime() */
  #include <sys/time.h>
#else
  /* for clock() */
  #include <time.h>
#endif

/* check for ctrl-c interrupt via utIsInterruptPending() */
#if defined (_WIN32)
  #include <windows.h>
#elif defined (__linux__)
  #include <unistd.h>
#endif

#include "mex.h"

#if __STDC_VERSION__ >= 199901L
/* C99 code */
#define C99
  #define STATIC static
  #define INLINE inline
#else
  #define STATIC
  #define INLINE
#endif

#ifdef _OPENMP
#include <omp.h>
#endif

/* Display constants, default is verbose */
#define DISPLAY_OFF 0
#define DISPLAY_XS 5
#define DISPLAY_ITER 10
#define DISPLAY_VERBOSE 255
#define DISPLAY_INTERVAL 30.0

extern int utIsInterruptPending();

/* if enabled, use 64-bit map instead of 32-bit map */
#define LSIZEREDUCE_64_BIT

/* if enabled, check for correctness and return error to matlab if check fails */
#undef  LSIZEREDUCE_USE_ASSERT


#ifdef LSIZEREDUCE_64_BIT
  #define GET_BIT(bitmap, index) get_bit_64(bitmap, index)
  #define SET_BIT(bitmap, index) set_bit_64(bitmap, index)
  #define UINT uint64_t
  #define INT int64_t
#else
  #define GET_BIT(bitmap, index) get_bit_32(bitmap, index)
  #define SET_BIT(bitmap, index) set_bit_32(bitmap, index)
  #define UINT uint32_t
  #define INT int32_t
#endif

#ifdef LSIZEREDUCE_USE_ASSERT
  #warning using assert
  static void assert_lsizereduce_fail(char *cond, const char *file, unsigned int line)
  {
    mexErrMsgIdAndTxt("gen_lattice_mex:assert", "assert failed in %s at line %d: %s", file, line, cond);
  }
  #define assert_lsizereduce(condition) {if (!(condition)) assert_lsizereduce_fail(__STRING(condition), __FILE__, __LINE__);}
#endif

/**
  * Returns current time in seconds for time measurements.
  * On Linux, the current system time (gettimeofday()) is used,
  * otherwise the processor time used by the current program (clock()).
  */
STATIC INLINE double getcurtime(void)
{
#ifdef __linux__
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return 1.0 * tv.tv_sec + 1e-6 * tv.tv_usec;
#else
  return (double) clock() / CLOCKS_PER_SEC;
#endif
}

/**
  * Calculates rem = a mod b, where 0 <= rem < b.
  */
STATIC INLINE UINT modl(int64_t a, INT b)
{
  INT rem = a % b;
  if (rem < 0)
    rem += b;
#ifdef LSIZEREDUCE_USE_ASSERT
  assert_lsizereduce(rem >= 0 && rem < b);
#endif

  return (UINT) rem;
}

#ifdef LSIZEREDUCE_64_BIT
/**
  * Gets bit at certain index from bitmap temp.
  * Uses 64-bit operations.
  */
STATIC INLINE uint64_t get_bit_64(uint64_t *temp, uint64_t index)
{
  uint64_t tindex = index >> 6U;
  uint8_t toffset = index & 0x3F;
  uint64_t tbit = 1UL << toffset;

  return temp[tindex] & tbit;
}

/**
  * Sets bit at certain index in bitmap temp.
  * Uses 64-bit operations.
  */
STATIC INLINE void set_bit_64(uint64_t *temp, uint64_t index)
{
  uint64_t tindex = index >> 6U;
  uint8_t toffset = index & 0x3F;
  uint64_t tbit = 1UL << toffset;

  temp[tindex] |= tbit;
}
#else
/**
  * Gets bit at certain index from bitmap temp.
  * Uses 32-bit operations.
  */
STATIC INLINE uint32_t get_bit_32(uint32_t *temp, uint32_t index)
{
  uint32_t tindex = index >> 5U;
  uint8_t toffset = index & 0x1F;
  uint32_t tbit = 1U << toffset;

  return temp[tindex] & tbit;
}

/**
  * Sets bit at certain index in bitmap temp.
  * Uses 32-bit operations.
  */
STATIC INLINE void set_bit_32(uint32_t *temp, uint32_t index)
{
  uint32_t tindex = index >> 5U;
  uint8_t toffset = index & 0x1F;
  uint32_t tbit = 1U << toffset;

  temp[tindex] |= tbit;
}
#endif

STATIC INLINE INT min_i(INT a, INT b)
{
  return a < b ? a : b;
}

STATIC INLINE INT max_i(INT a, INT b)
{
  return a >= b ? a : b;
}

STATIC INLINE int32_t min_i32(int32_t a, int32_t b)
{
  return a < b ? a : b;
}

static void flushMatlab(void)
{
  fflush(NULL);
/*  mexEvalString("drawnow;"); */
}

/* Adapted from http://www.algorithmist.com/index.php/Prime_Sieve_of_Eratosthenes.c */
STATIC INLINE UINT prime_sieve_get(UINT *sieve, UINT b)
{
#ifdef LSIZEREDUCE_64_BIT
  return (sieve[(b)>>6]>>((b)&63))&1UL;
#else
  return (sieve[(b)>>5]>>((b)&31))&1;
#endif
}
static void prime_sieve(UINT *sieve, UINT P1, UINT P2, UINT P3)
{
  UINT i, j, k;
  memset(sieve, 0, P1*sizeof(UINT));
  for (k = 1; k <= P3; k++)
    if (prime_sieve_get(sieve, k)==0)
      for(j=2*k+1,i=2*k*(k+1);i<P2;i+=j)
#ifdef LSIZEREDUCE_64_BIT
        sieve[i>>6]|=1UL<<(i&63);
#else
        sieve[i>>5]|=1<<(i&31);
#endif
}
STATIC INLINE INT isprime(UINT *sieve, INT p)
{
  return p==2 || (p>2 && (p&1)==1 && (prime_sieve_get(sieve, (p-1)>>1)==0));
}

static int test_reconstruction_property(
  UINT *bitmap,
  const INT bitmap_size,
  INT len_freq,
  int64_t *tempsp,
  INT curM
)
{
  INT k;

  memset(bitmap, 0, bitmap_size);
  for (k = 0; k < len_freq; k++)
  {
    UINT index;
    index = modl(tempsp[k], curM);
        
    if (GET_BIT(bitmap, index) != 0)
      break;
    SET_BIT(bitmap, index);
  }

  if (k == len_freq)
    return 1;

  return 0;
}

/**
  */
static INT lattice_size_reduce_bisect_only(int d, INT *z, INT *freq, INT len_freq, INT M_input, int successors, int prime, int nthreads, int display)
{
  INT t, k;

#ifdef LSIZEREDUCE_64_BIT
  const INT bitmap_length = 1+M_input/(1U<<6);
  const INT bitmap_size = bitmap_length * sizeof(uint64_t);
#else
  const INT bitmap_length = 1+M_input/(1U<<5);
  const INT bitmap_size = bitmap_length * sizeof(uint32_t);
#endif
  int64_t *tempsp = (int64_t *) mxMalloc(len_freq * sizeof(int64_t));
  UINT *sieve;
  INT lastSuccess = -1;
  double time0, time1, time2;
  UINT *candidates;
  INT len_candidates = 0;
  INT j = 0, j_u = 0, j_o = 0;
  bool first_loop = true;


  time0 = getcurtime();

  if (prime == 1)
  {
    INT curM;
    if (display >= DISPLAY_VERBOSE)
    {
      printf("Searching for prime numbers ...\n");
      flushMatlab();
    }

    sieve = (UINT *) mxMalloc((M_input/64 + 1) * sizeof(UINT));
#ifdef LSIZEREDUCE_64_BIT
    prime_sieve(sieve, M_input/64 + 1, llrint(ceil(M_input/2.0)), llrint(ceil(ceil(sqrt((double) M_input))/2.0)));
#else
    prime_sieve(sieve, M_input/64 + 1, lrint(ceil(M_input/2.0)), lrint(ceil(ceil(sqrt((double) M_input))/2.0)));
#endif

    for (curM = len_freq; curM <= M_input; curM++)
      if (isprime(sieve, curM) == 1)
        len_candidates++;

    if (display >= DISPLAY_VERBOSE)
    {
#ifdef LSIZEREDUCE_64_BIT
      printf("len_candidates = %lld\n", len_candidates);
#else
      printf("len_candidates = %d\n", len_candidates);
#endif
      flushMatlab();
    }

    if (len_candidates < 1)
    {
      if (display > DISPLAY_OFF)
      {
        printf("len_candidates is zero\n");
        flushMatlab();
      }
      mxFree(tempsp);
      mxFree(sieve);
      return 0;
    }

    candidates = (UINT *) mxMalloc(len_candidates * sizeof(UINT));
    k = 0;
    for (curM = len_freq; curM <= M_input; curM++)
      if (isprime(sieve, curM) == 1)
      {
        candidates[k] = curM;
        k++;
      }
  }

  if (display >= DISPLAY_VERBOSE)
  {
    printf("Calculating inner product freq * z ...\n");
    flushMatlab();
  }

  for (k = 0; k < len_freq; k++)
  {
    tempsp[k] = freq[d*k] * z[0];
    for (t = 1; t < d; t++)
      tempsp[k] += freq[d*k+t] * z[t];
  }

  if (display >= DISPLAY_VERBOSE)
  {
    printf("Starting bisection search for smaller lattice size M ...\n");
    flushMatlab();
  }

  time1 = time0;

    if (prime==1)
    {
      j_u = len_candidates;
      j_o = 0;
    }
    else
    {
      j_u = M_input+1;
      j_o = len_freq;
    }
    j = (j_u+j_o)/2;
    while(first_loop || (j > j_o && j < j_u))
    {
      INT r, r_max;
      volatile bool interrupted = false;
      volatile bool found = false;

      first_loop = false;

      if (j_u - j_o < successors)
        j = j_o;

/*
    if (display >= DISPLAY_VERBOSE)
    {
      if (prime == 1)
#ifdef LSIZEREDUCE_64_BIT
        printf("j %lld, j_o %lld, j_u %lld | c_j %lld, c_j_o %lld, c_j_u %lld\n", j, j_o, j_u, candidates[j], candidates[j_o], candidates[j_u]);
#else
        printf("j %d, j_o %d, j_u %d | c_j %d, c_j_o %d, c_j_u %d\n", j, j_o, j_u, candidates[j], candidates[j_o], candidates[j_u]);
#endif
      else 
#ifdef LSIZEREDUCE_64_BIT
        printf("j %lld, j_o %lld, j_u %lld\n", j, j_o, j_u);
#else
        printf("j %d, j_o %d, j_u %d\n", j, j_o, j_u);
#endif
      flushMatlab();
    }
*/
      r_max = min_i(j_u-1, j+successors);

#pragma omp parallel shared(found,interrupted) num_threads(nthreads)
{
#ifdef LSIZEREDUCE_64_BIT
      uint64_t *bitmap;
      #pragma omp critical
      {
        bitmap = (uint64_t *) mxMalloc(bitmap_size);
      }
#else
      uint32_t *bitmap;
      #pragma omp critical
      {
        bitmap = (uint32_t *) mxMalloc(bitmap_size);
      }
#endif

      #pragma omp for schedule(dynamic,1)
      for (r = j; r <= r_max; r++)
      {
        INT cur_bitmap_size;

        if (found || interrupted)
          continue;
#ifdef _OPENMP
        if (omp_get_thread_num() == 0)
#endif
        {
          if (utIsInterruptPending() != 0) {
	   #pragma omp critical
	   {
            mexWarnMsgTxt("CTRL-C detected. Aborting...");
            flushMatlab();

            interrupted = true;
	   }
            continue;
          }
        }

#ifdef _OPENMP
       if (omp_get_thread_num() == 0)
#endif
       {
        if (display >= DISPLAY_VERBOSE || display == DISPLAY_XS)
        {
          time2 = getcurtime();
          if (time2 - time1 > DISPLAY_INTERVAL)
          {
	   #pragma omp critical
	   {
            time1 = time2;
            if (lastSuccess == -1)
#ifdef LSIZEREDUCE_64_BIT
              printf("curr candidate: %lld\n", prime==1 ? candidates[r] : r);
#else
              printf("curr candidate: %d\n", prime==1 ? candidates[r] : r);
#endif
            else
#ifdef LSIZEREDUCE_64_BIT
              printf("last success: %lld, curr candidate: %lld\n", prime==1 ? candidates[lastSuccess] : lastSuccess, prime==1 ? candidates[r] : r);
#else
              printf("last success: %d, curr candidate: %d\n", prime==1 ? candidates[lastSuccess] : lastSuccess, prime==1 ? candidates[r] : r);
#endif
            flushMatlab();
	   }
          }
        }
       }
if (prime==1 && r >= len_candidates)
{
  #pragma omp critical
  {
    if (display > DISPLAY_OFF)
    {
#ifdef LSIZEREDUCE_64_BIT
      printf("error: r=%lld >= len_candidates=%lld\n", r, len_candidates);
#else
      printf("error: r=%d >= len_candidates=%d\n", r, len_candidates);
#endif
      flushMatlab();
    }
	    interrupted = true;
  }
	    continue;
}

#ifdef LSIZEREDUCE_64_BIT
        cur_bitmap_size = (prime == 1 ? (1+candidates[r]/(1U<<6)) * sizeof(uint64_t) : (1+r/(1U<<6)) * sizeof(uint64_t));
#else
        cur_bitmap_size = (prime == 1 ? (1+candidates[r]/(1U<<5)) * sizeof(uint32_t) : (1+r/(1U<<5)) * sizeof(uint32_t));
#endif

        if (test_reconstruction_property(bitmap, cur_bitmap_size, len_freq, tempsp, prime==1 ? candidates[r] : r) == 1)
	{
          #pragma omp critical
          {
            if (found)
            {
              if (r < lastSuccess)
                lastSuccess = r;
	    }
	    else
              lastSuccess = r;

	    found = true;

            if (display >= DISPLAY_ITER)
            {
#ifdef _OPENMP
  #ifdef LSIZEREDUCE_64_BIT
              printf("thread %d found lattice size: %lld\n", omp_get_thread_num(), prime==1 ? candidates[r] : r);
  #else
              printf("thread %d found lattice size: %d\n", omp_get_thread_num(), prime==1 ? candidates[r] : r);
  #endif
#else
  #ifdef LSIZEREDUCE_64_BIT
              printf("found lattice size: %lld\n", prime==1 ? candidates[r] : r);
  #else
              printf("found lattice size: %d\n", prime==1 ? candidates[r] : r);
  #endif
#endif
              flushMatlab();
              time1 = getcurtime();
            }
	  }
	}
      } /* end for(r) */

      #pragma omp critical
      {
        mxFree(bitmap);
      }
}

      if (interrupted)
      {
        mxFree(tempsp);
        if (prime == 1)
        {
          mxFree(sieve);
          mxFree(candidates);
        }
	return -1;
      }

      if (found)
        j_u = j;

      if (j_u != j)
        j_o = j + successors;
      j = (j_u+j_o)/2;
    }

if (prime==1 && j_o >= len_candidates)
{
  if (display > DISPLAY_OFF)
  {
#ifdef LSIZEREDUCE_64_BIT
    printf("bisection did not find smaller lattice: j_o=%lld >= len_candidates=%lld\n", j_o, len_candidates);
#else
    printf("bisection did not find smaller lattice: j_o=%d >= len_candidates=%d\n", j_o, len_candidates);
#endif
    flushMatlab();
  }
    mxFree(tempsp);
    mxFree(sieve);
    mxFree(candidates);
    return 0;
}

  if (prime == 1 && lastSuccess >= 0)
    lastSuccess = candidates[lastSuccess]; 

  mxFree(tempsp);
  if (prime == 1)
  {
    mxFree(sieve);
    mxFree(candidates);
  }

  if (display >= DISPLAY_ITER)
  {
    printf("Time: %.3e\n", getcurtime() - time0);
  }

  return lastSuccess > 0 ? lastSuccess : 0;
}

void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[])
{
  double *dfreq;
  INT nfreq;
  int d;
  INT M;
  INT outM;
  INT l, t;
  INT *freq;
  INT *z;
  int display = DISPLAY_VERBOSE;
  int prime = 0;
  double *zin;
  int lzin;
  int successors = 0;
#ifdef _OPENMP
  int nthreads = omp_get_max_threads();
#else
  int nthreads = 1;
#endif

  if(nrhs < 3)
    mexErrMsgIdAndTxt("gen_lattice_mex:nrhs",
                      "Three inputs required. Usage: M = lattice_size_reduce_mex(frequency_index_set, lattice_size_M, z)");

  if(nlhs!=1)
    mexErrMsgIdAndTxt("gen_lattice_mex:nlhs",
                      "One output required. Usage: lattice_size_reduce_mex(frequency_index_set, lattice_size_M, z)");

  if(mxGetNumberOfElements(prhs[1]) != 1 )
    mexErrMsgIdAndTxt("gen_lattice_mex:M_not_scalar",
                      "Input lattice size M (second argument) must be a scalar.");

  nfreq = mxGetM(prhs[0]);
  d = mxGetN(prhs[0]);
#ifdef LSIZEREDUCE_64_BIT
  if (mxGetScalar(prhs[1]) > (double) (1UL<<53))
    mexErrMsgIdAndTxt("gen_lattice_mex:M_too_large",
                      "Input lattice size M (second argument) must not be larger than 2^53.");
#else
  if (mxGetScalar(prhs[1]) > (double) (1U<<31))
    mexErrMsgIdAndTxt("gen_lattice_mex:M_too_large",
                      "Input lattice size M (second argument) must not be larger than 2^31.");
#endif

#ifdef LSIZEREDUCE_64_BIT
  M = llrint(mxGetScalar(prhs[1]));
#else
  M = lrint(mxGetScalar(prhs[1]));
#endif

  if (nrhs > 3 && mxIsStruct(prhs[3]))
  {
    int index = mxIsStruct(prhs[2]) ? 2 : 3;
    char str[256];
    mxArray *field;

    field = mxGetField(prhs[index], 0, "display");
    if (field == NULL)
      field = mxGetField(prhs[index], 0, "Display");
    if (field != NULL && mxGetString(field, str, 255) == 0)
    {
      if (strcmp(str, "off") == 0 || strcmp(str, "Off") == 0
        ||strcmp(str, "none") == 0 || strcmp(str, "None") == 0)
        display = DISPLAY_OFF;
      else if (strcmp(str, "iter") == 0 || strcmp(str, "Iter") == 0)
        display = DISPLAY_ITER;
      else if (strcmp(str, "xs") == 0 || strcmp(str, "Xs") == 0)
        display = DISPLAY_XS;
    }

    field = mxGetField(prhs[index], 0, "prime");
    if (field == NULL)
      field = mxGetField(prhs[index], 0, "Prime");
    if (field != NULL && lrint(mxGetScalar(field)) == 1)
      prime = 1;

    field = mxGetField(prhs[index], 0, "Bisection_Successors");
    if (field == NULL)
      field = mxGetField(prhs[index], 0, "bisection_successors");
      if (field != NULL)
        successors = lrint(mxGetScalar(field));

    field = mxGetField(prhs[index], 0, "nthreads");
    if (field != NULL)
    {
#ifdef _OPENMP
      nthreads = min_i32(lrint(mxGetScalar(field)), omp_get_max_threads());
#else
      nthreads = min_i32(lrint(mxGetScalar(field)), 1);
#endif

      if (nthreads < 1)
        nthreads = 1;
    }
  }

  if (display > DISPLAY_OFF)
  {
#ifdef LSIZEREDUCE_64_BIT
    printf("d = %d, card_I = %lld, lattice_size_M = %lld, successors = %d, prime = %d, nthreads = %d, display = ", d, nfreq, M, successors, prime, nthreads);
#else
    printf("d = %d, card_I = %d, lattice_size_M = %d, successors = %d, prime = %d, nthreads = %d, display = ", d, nfreq, M, successors, prime, nthreads);
#endif
    if (display == DISPLAY_ITER)
      printf("iter");
    else if (display == DISPLAY_VERBOSE)
      printf("verbose");
    else if (display == DISPLAY_XS)
      printf("xs");
    printf("\n");
    flushMatlab();
  }

  if (M < nfreq)
    mexErrMsgIdAndTxt("gen_lattice_mex:M_less_than_card",
                      "Input lattice_size_M is smaller than cardinality of frequency index set.");

  dfreq = mxGetPr(prhs[0]);

  z = (INT *) mxMalloc(d*sizeof(INT));

  lzin = mxGetN(prhs[2]);
  if (mxIsStruct(prhs[2]) || lzin != d)
    mexErrMsgIdAndTxt("gen_lattice_mex:invalid_z",
                      "Input generating vector z (third argument) must be row vector of length d");

  zin = mxGetPr(prhs[2]);
  for (t = 0; t < d; t++)
#ifdef LSIZEREDUCE_64_BIT
    z[t] = llrint(zin[t]);
#else
    z[t] = lrint(zin[t]);
#endif

#ifdef LSIZEREDUCE_64_BIT
  if (prime == 1 && M > (1UL<<36))
    mexErrMsgIdAndTxt("gen_lattice_mex:M_too_large",
                      "Input lattice size M (second argument) must not be larger than 2^36.");
#endif

  if (display > DISPLAY_OFF)
  {
    printf("z: ");
    for (t = 0; t < d; t++)
#ifdef LSIZEREDUCE_64_BIT
      printf("%lld ", z[t]);
#else
      printf("%d ", z[t]);
#endif
    printf("\n");
  }

  freq = (INT *) mxMalloc(d*nfreq*sizeof(INT));
  for (l = 0; l < nfreq; l++)
    for (t = 0; t < d; t++)
#ifdef LSIZEREDUCE_64_BIT
      freq[d*l+t] = llrint(dfreq[nfreq*t+l]);
#else
      freq[d*l+t] = lrint(dfreq[nfreq*t+l]);
#endif

  if (display >= DISPLAY_VERBOSE)
  {
#ifdef LSIZEREDUCE_64_BIT
      printf("64-bit map");
#else
      printf("32-bit map");
#endif
      printf("\n");
      flushMatlab();
  }

  outM = lattice_size_reduce_bisect_only(d, z, freq, nfreq, M, successors, prime, nthreads, display);

  if (outM == 0)
    outM = M;

  if (outM > 0)
  {
    double *retM;
    plhs[0] = mxCreateDoubleMatrix(1,1,mxREAL);
    retM = mxGetPr(plhs[0]);
    retM[0] = outM;

    if (display > DISPLAY_OFF)
    {
#ifdef LSIZEREDUCE_64_BIT
      printf("M = %lld\n", outM);
#else
      printf("M = %d\n", outM);
#endif
      flushMatlab();
    }
  }

  mxFree(freq);
  mxFree(z);

  if (outM == -1)
  {
    mexErrMsgIdAndTxt("gen_lattice_mex:ctrl-c",
                      "CTRL-C pressed. Aborted.");
  }
}

